import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:rx_property/rx_property.dart';

extension RxFlutterExtension<T> on Reactive<T> {
  Widget builder(
      {@required RxDataWidgetBuilder<T> data,
      @required RxErrorWidgetBuilder<T> error,
      @required RxLoadingWidgetBuilder<T> loading}) {
    return ReactiveBuilder<T>(
      builder: data,
      errorBuilder: error,
      loadingBuilder: loading,
      reactive: this,
    );
  }
}

typedef RxDataWidgetBuilder<T> = Widget Function(BuildContext context, T data);
typedef RxErrorWidgetBuilder<T> = Widget Function(
    BuildContext context, Object error);
typedef RxLoadingWidgetBuilder<T> = Widget Function(BuildContext context);

class ReactiveBuilder<T> extends StatelessWidget {
  final Reactive<T> reactive;
  final RxDataWidgetBuilder<T> builder;
  final RxErrorWidgetBuilder<T> errorBuilder;
  final RxLoadingWidgetBuilder<T> loadingBuilder;
  const ReactiveBuilder(
      {Key key,
      @required this.reactive,
      @required this.builder,
      @required this.errorBuilder,
      @required this.loadingBuilder})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return RxStreamBuilder<T>(
      stream: reactive,
      builder: builder,
      errorBuilder: errorBuilder,
      loadingBuilder: loadingBuilder,
    );
  }
}

class RxStreamBuilder<T> extends StatelessWidget {
  final Stream<T> stream;
  final RxDataWidgetBuilder<T> builder;
  final RxErrorWidgetBuilder<T> errorBuilder;
  final RxLoadingWidgetBuilder<T> loadingBuilder;

  const RxStreamBuilder(
      {Key key,
      @required this.stream,
      @required this.builder,
      @required this.errorBuilder,
      @required this.loadingBuilder})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return StreamBuilder<T>(
      stream: stream,
      builder: (context, snapshot) {
        if (snapshot.hasError) return errorBuilder(context, snapshot.error);
        if (snapshot.hasData) return builder(context, snapshot.data);
        return loadingBuilder(context);
      },
    );
  }
}
